@@ -7,7 +7,7 @@
 # GNU General Public License version 2 or any later version.
 
 import os, copy
-from mercurial import match, patch, scmutil, error, ui, util
+from mercurial import match, patch, error, ui, util, pathutil, context
 from mercurial.i18n import _
 from mercurial.node import hex, nullid
 from common import ErrorResponse
@@ -138,6 +138,9 @@ def _siblings(siblings=[], hiderev=None):
         yield d
 
 def parents(ctx, hide=None):
+    if (isinstance(ctx, context.basefilectx) and
+        ctx.changectx().rev() != ctx.linkrev()):
+        return _siblings([ctx._repo[ctx.linkrev()]], hide)
     return _siblings(ctx.parents(), hide)
 
 def children(ctx, hide=None):
@@ -146,7 +149,7 @@ def children(ctx, hide=None):
 def renamelink(fctx):
     r = fctx.renamed()
     if r:
-        return [dict(file=r[0], node=hex(r[1]))]
+        return [{'file': r[0], 'node': hex(r[1])}]
     return []
 
 def nodetagsdict(repo, node):
@@ -196,7 +199,7 @@ def showbookmark(repo, tmpl, t1, node=nullid, **args):
 
 def cleanpath(repo, path):
     path = path.lstrip('/')
-    return scmutil.canonpath(repo.root, '', path)
+    return pathutil.canonpath(repo.root, '', path)
 
 def changeidctx (repo, changeid):
     try:
@@ -246,6 +249,35 @@ def filectx(repo, req):
 
     return fctx
 
+def changelistentry(web, ctx, tmpl):
+    '''Obtain a dictionary to be used for entries in a changelist.
+
+    This function is called when producing items for the "entries" list passed
+    to the "shortlog" and "changelog" templates.
+    '''
+    repo = web.repo
+    rev = ctx.rev()
+    n = ctx.node()
+    showtags = showtag(repo, tmpl, 'changelogtag', n)
+    files = listfilediffs(tmpl, ctx.files(), n, web.maxfiles)
+
+    return {
+        "author": ctx.user(),
+        "parent": parents(ctx, rev - 1),
+        "child": children(ctx, rev + 1),
+        "changelogtag": showtags,
+        "desc": ctx.description(),
+        "extra": ctx.extra(),
+        "date": ctx.date(),
+        "files": files,
+        "rev": rev,
+        "node": hex(n),
+        "tags": nodetagsdict(repo, n),
+        "bookmarks": nodebookmarksdict(repo, n),
+        "inbranch": nodeinbranch(repo, ctx),
+        "branches": nodebranchdict(repo, ctx)
+    }
+
 def listfilediffs(tmpl, files, node, max):
     for f in files[:max]:
         yield tmpl('filedifflink', node=hex(node), file=f)
